#include <iostream>
#include <limits>
#include "dynamic.h"
#include "global.h"

using namespace std;

struct Solution {
		unsigned long long X;
		int c;
		int M;
};

static Solution ***map;
static unsigned long long step;

Solution KNAP(unsigned long long V, int M, int last);
Solution trivialKNAP(unsigned long long V, int M);
unsigned long long reset(unsigned long long V, int i);
unsigned long long set(unsigned long long V, int i);
bool isTrivial (unsigned long long V, int M);

int dynamic(bool print) {
		map = new Solution **[n];
		for(int i = 0; i < n; i++) {
				map[i] = new Solution *[maxWeight];
				for(int j = maxWeight-1; j >= 0; j--) {
						map[i][j] = NULL;
				}
		}

		step = 0;
		Solution s = KNAP(getLongLong(n) - 1, maxWeight, n-1);

		for(int i = 0; i < n; i++) {
				for(int j = maxWeight-1; j >= 0; j--) {
						delete map[i][j];
				}
				delete [] map[i];
		}
		delete map;

		if(print) printSolution(s.X, s.c, step);

		return s.c;
}

Solution KNAP(unsigned long long V, int M, int last) {
		step++;
    if(isTrivial(V, M)) return trivialKNAP(V, M);
		if(map[last][M-1]) {
				Solution res = {map[last][M-1]->X, map[last][M-1]->c, map[last][M-1]->M};
				return res;
		}

		Solution s1 = KNAP(reset(V, last), M, last-1);
		Solution s2 = KNAP(reset(V, last), M-items[last][0], last-1);

		s2.X = set(s2.X, last); s2.c += items[last][1]; s2.M += items[last][0];
		s1.X = reset(s1.X, last);

		if((s2.c > s1.c) && (s2.M <= maxWeight)) {
				map[last][M-1] = new Solution(s2);
				return s2;
		} else {
				map[last][M-1] = new Solution(s1);
				return s1;
		}
}

unsigned long long reset(unsigned long long V, int i) {
		return V & ~getLongLong(i);
}

unsigned long long set(unsigned long long V, int i) {
		return V | getLongLong(i);
}

bool isTrivial (unsigned long long V, int M) {
    return (V == 0) || (M <= 0);
}

Solution trivialKNAP(unsigned long long V, int M) {
		if(M < 0) {
				Solution res = {0, INT_MIN, 0};
				return res;
		} else {
				Solution res = {0, 0, 0};
				return res;
		}
}
